import numpy as np
import pandas as pd
import pandas_market_calendars as pmc
import pytest

from tradingo import signals
from tradingo.api import Tradingo

UNIVERSE = "test-universe"
PROVIDER = "test-provider"


def test_buffer_signal(tradingo: Tradingo):
    signals.buffered(
        signal=tradingo.portfolio.model.raw.shares(),
        buffer_width=0.5,
    )


@pytest.mark.parametrize(
    "ask_close, bid_close, close_offset_periods",
    [
        (
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            0,
        ),
        (
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            pd.DataFrame(
                np.cumprod(1 + np.random.normal(0, 0.009, 167905)),
                index=pd.bdate_range(
                    start="2020-01-01 00:00:00+00:00",
                    end="2024-10-15 18:30:00+00:00",
                    freq=pd.offsets.Minute(15),
                ),
                columns=["ABCD"],
            ),
            3,
        ),
    ],
)
def test_intraday_momentum(
    ask_close,
    bid_close,
    close_offset_periods,
):
    result = pd.concat(
        signals.intraday_momentum(
            ask_close=ask_close,
            bid_close=bid_close,
            close_offset_periods=close_offset_periods,
            calendar="NYSE",
        ),
        axis=1,
        keys=(
            "intraday_momentum",
            "z_score",
            "short_vol",
            "long_vol",
            "previous_close_px",
        ),
    )

    cal = pmc.get_calendar("NYSE")
    schedule = cal.schedule(start_date="2024-10-14", end_date="2024-10-14")
    trading_index = pmc.date_range(schedule, frequency="15min")

    subset = result.loc[trading_index]

    assert subset[-(close_offset_periods + 1) :]["intraday_momentum"][
        "ABCD"
    ].to_list() == [0 for _ in range(0, 1 + close_offset_periods)]


if __name__ == "__main__":
    pytest.main([__file__])



import pytest

from tradingo.sampling.yf import currency_to_symbol, symbol_to_currency


def test_handle_currency_ticker_valid_ticker():
    symbols = "USDJPY=X"
    result = symbol_to_currency(symbols)
    assert (
        result == "USDJPY"
    ), "The ticker should be correctly converted to currency pair."


def test_handle_currency():
    ticker = "GBPUSD"
    result = currency_to_symbol(ticker)
    assert result == "GBPUSD=X", "The ticker is a currency pair, whould end by =X."


def test_handle_not_a_currency():
    ticker = "NOTCCY"
    result = currency_to_symbol(ticker)
    assert result == "NOTCCY", "The ticker is not a currency pair."


if __name__ == "__main__":
    pytest.main([__file__])



"""
Factor model and risk decomposition.
"""

__all__ = [
    "factor_betas",
    "idiosyncratic_alpha",
    "idiosyncratic_residuals",
    "factor_model",
]

import logging

import pandas as pd

from tradingo.analytics import cov, expectation

logger = logging.getLogger(__name__)


def factor_betas(
    asset_returns: pd.DataFrame,
    factor_returns: pd.DataFrame,
    how: str = "exponential",
    **kwargs,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Regress asset vs factor returns (dependent vs independent variables)
    to calculate alpha (residuals) and beta (loadings).

    :param asset_returns: Asset returns DataFrame.
    :param factor_returns: Factor returns DataFrame.
    :param how: Method of calculating covariance, either "exponential" or "rolling",
        by default "exponential".
    :param **kwargs: Additional arguments for the ewm (exponentially weighted mean) method.
        For rolling covariance, the window size must be specified in kwargs.
        For example, `window=10` for a 10-period rolling covariance.
        For exponential covariance, the halflife can be specified in kwargs.
        For example, `halflife=10` for a 10-period halflife.

    :returns: tuple with betas, and factor covariances.
    """

    if set(asset_returns.columns).intersection(factor_returns.columns):
        logger.warning("Asset returns and factor returns have overlapping columns. ")
    if asset_returns.index.freq != factor_returns.index.freq:
        raise ValueError("Asset and factor returns must have the same frequency.")

    start = asset_returns.first_valid_index()
    returns = pd.concat((asset_returns.loc[start:], factor_returns.loc[start:]), axis=1)

    cov_mtx = cov(
        returns,
        annualisation=1,
        how=how,
        **kwargs,
    ).reindex(asset_returns.index, level=0)

    factor_covs = (
        cov_mtx.loc[(slice(None), factor_returns.columns), factor_returns.columns]
        .sort_index(level=0)
        .reindex(asset_returns.index, level=0, axis=0)
    )

    # cov to beta
    betas = (
        pd.concat(
            (
                cov_mtx.loc[(slice(None), asset_returns.columns), factor]
                .sort_index(level=0)
                .unstack(level=1)
                .divide(cov_mtx.loc[(slice(None), factor), factor].to_numpy(), axis=0)
                .stack(dropna=False)
                for factor in factor_returns.columns
            ),
            axis=1,
            keys=factor_returns.columns,
        )
        .unstack(level=1)
        .stack(level=0)
    ).reindex(index=factor_returns.columns, level=1)

    return betas, factor_covs


def idiosyncratic_alpha(
    asset_returns: pd.DataFrame,
    factor_returns: pd.DataFrame,
    betas: pd.DataFrame = None,
    how: str = "exponential",
    **kwargs,
) -> pd.DataFrame:
    """
    Calculate idiosyncratic alphas from asset returns and factor returns.
    Alphas are calculated as the difference between asset returns and
    the product of betas and factor returns.

    :param asset_returns: Asset returns DataFrame.
    :param factor_returns: Factor returns DataFrame.
    :param betas: Betas DataFrame, by default None.
    :param how: Method of calculating covariance, either "exponential" or "rolling",
        by default "exponential".
    :param **kwargs: Additional arguments for the ewm (exponentially weighted mean) method.
        For rolling covariance, the window size must be specified in kwargs.
        For example, `window=10` for a 10-period rolling covariance.
        For exponential covariance, the halflife can be specified in kwargs.
        For example, `halflife=10` for a 10-period halflife.

    :returns: tuple with alphas and mean returns dataframes
    """

    if betas is None:
        betas, _ = factor_betas(
            asset_returns,
            factor_returns,
            how=how,
            **kwargs,
        )
    elif not betas.index.get_level_values(0).difference(asset_returns.index).empty:
        raise ValueError(
            "Asset returns and betas must have the same index. "
            f"Got {asset_returns.index} and {betas.index}"
        )

    if set(asset_returns.columns).intersection(factor_returns.columns):
        logger.warning("Asset returns and factor returns have overlapping columns. ")
    if asset_returns.index.freq != factor_returns.index.freq:
        raise ValueError("Asset and factor returns must have the same frequency.")

    start = asset_returns.first_valid_index()
    returns = pd.concat((asset_returns.loc[start:], factor_returns.loc[start:]), axis=1)

    # calculate alpha = E[asset] - beta * E[factor]
    means = expectation(
        returns,
        annualisation=1,
        how=how,
        **kwargs,
    ).dropna(axis=0, how="all")

    index = means.index.intersection(
        betas.index.get_level_values(0).unique().sort_values()
    )
    betas = betas.reindex(index=index, level=0)
    means = means.reindex(index=index, level=0)

    alphas = (
        pd.concat(
            (
                means.loc[:, asset_returns.columns]
                .sort_index(level=0)
                .subtract(
                    betas.loc[(slice(None), factor), asset_returns.columns]
                    .reindex(
                        pd.MultiIndex.from_product([index, [factor]]), fill_value=None
                    )
                    .mul(means.loc[:, [factor]].to_numpy()),
                    axis=0,
                )
                for factor in factor_returns.columns
            ),
            axis=0,
        )
        .sort_index()
        .reindex(index=factor_returns.columns, level=1)
        .dropna(how="all")
    )

    return alphas, means


def idiosyncratic_residuals(
    asset_returns: pd.DataFrame,
    factor_returns: pd.DataFrame,
    betas: pd.DataFrame,
    alphas: pd.DataFrame,
) -> pd.DataFrame:
    """
    Calculate regression residuals: observed return minus model-predicted return.
    """

    idx = asset_returns.index.intersection(betas.index.get_level_values(0))
    asset_returns = asset_returns.reindex(idx)
    factor_returns = factor_returns.reindex(idx)
    betas = betas.reindex(index=idx, level=0)
    alphas = alphas.reindex(index=idx, level=0)

    predicted_returns = pd.concat(
        [
            betas.loc[(slice(None), factor), :]
            .multiply(factor_returns[factor], axis=0)
            .add(alphas.loc[(slice(None), factor), :])
            for factor in factor_returns.columns
        ],
        axis=0,
    ).sort_index()

    residuals = asset_returns - predicted_returns
    return residuals


def factor_model(
    asset_returns: pd.DataFrame,
    factor_returns: pd.DataFrame,
    how: str = "exponential",
    **kwargs,
):
    """
    end-to-end multi-factor model regression, including confidence intervals.

    :param asset_returns: Asset returns DataFrame.
    :param factor_returns: Factor returns DataFrame.
    :param how: Method of calculating covariance, either "exponential" or "rolling",
        by default "exponential".
    :param **kwargs: Additional arguments for the ewm (exponentially weighted mean) method.
        For rolling covariance, the window size must be specified in kwargs.
        For example, `window=10` for a 10-period rolling covariance.
        For exponential covariance, the halflife can be specified in kwargs.
        For example, `halflife=10` for a 10-period halflife.

    :returns: tuple with alphas, betas, and their standard error dataframes.

    """

    betas, covs_factors = factor_betas(
        asset_returns,
        factor_returns,
        how=how,
        **kwargs,
    )

    alphas, means_returns = idiosyncratic_alpha(
        asset_returns,
        factor_returns,
        **kwargs,
    )

    # model confidence
    residuals = idiosyncratic_residuals(
        asset_returns,
        factor_returns,
        betas,
        alphas,
    )

    cov_mtx = (
        cov(
            residuals.unstack(),
            annualisation=1,
            how=how,
            **kwargs,
        )
        .unstack(level=(1, 2))
        .reindex(asset_returns.index, level=0)
    )
    cov_residuals = (
        cov_mtx[[idx for idx in cov_mtx.columns if idx[:2] == idx[2:]]]
        .droplevel((0, 1), axis=1)
        .stack()
    )

    covs_factors = covs_factors.unstack(fill_value=None)
    covs_factors = covs_factors[
        [c for c in covs_factors.columns if c[0] == c[1]]
    ].droplevel(0, axis=1)

    # alpha and beta confidence is measured through standard errors:
    # se_betas = std_residuals / std_factors
    # se_alphas = std_residuals * sqrt(1/n + mean_factors**2 / var_factors)

    se_betas = pd.concat(
        [
            cov_residuals.loc[(slice(None), factor), :].divide(
                covs_factors[factor], axis=0
            )
            ** 0.5
            for factor in factor_returns.columns
        ],
        axis=0,
    ).sort_index()

    means_factors = means_returns[factor_returns.columns]
    se_alphas = pd.concat(
        [
            cov_residuals.loc[(slice(None), factor), :].multiply(
                (1.0 / len(means_factors[factor]))
                + (means_factors[factor] ** 2).div(covs_factors[factor]),
                axis=0,
            )
            ** 0.5
            for factor in factor_returns.columns
        ],
        axis=0,
    ).sort_index()

    return alphas, betas, se_alphas, se_betas



"""Covariance calculation module."""

from typing import Callable

import numpy as np
import pandas as pd

from .expectation import safe_dataframe_window


def _kronecker_product(dataframe: pd.DataFrame) -> pd.DataFrame:
    """row-by-row outer product"""
    return pd.DataFrame(
        dataframe.apply(lambda x: np.kron(x, x), axis=1, result_type="expand"),
        index=dataframe.index,
        columns=pd.MultiIndex.from_product([dataframe.columns, dataframe.columns]),
    )


def cov(
    dataframe: pd.DataFrame,
    annualisation: int = 260,
    how: str = "ewm",
    demean: bool = True,
    **kwargs,
) -> pd.DataFrame:
    """
    Calculate the covariance of a given DataFrame.

    :param dataframe: DataFrame containing the data.
    :param annualisation: The annualisation factor, by default 260.
    :param how: A DataFrame method which returns a BaseWindow [pandas.core.window.rolling]
        for calculating covariance: { ewm | rolling | expanding }
    :param demean: Whether to demean the data, by default True.
            - True -> E[(X - E[X]) * (Y - E[Y])]
            - False -> E[X * Y]
        When True, it is the covariance.
        When False, we use Kronecker's product (row-wise pairs), from which we calc the mean.
        The result is the second moments, useful for raw signal magnitude, co-movement
        including mean shifts, or feeding into models where you handle the mean explicitly.
        NOTE: because of returns being "somewhere around zero", this '(co)variance'
        accounts for the implicit bias within a timeseries (e.g. a typical long position)

    :param **kwargs: Additional arguments for rolling/ewm/... pandas methods.
        For rolling covariance, the window size must be specified in kwargs.
        For example, `window=10` for a 10-period rolling covariance.
        For exponential covariance, the halflife can be specified in kwargs.
        For example, `halflife=10` for a 10-period halflife.

    :returns: The covariance of the data.
    """

    dataframe = dataframe if demean else _kronecker_product(dataframe)

    agg_buffer = safe_dataframe_window(dataframe, how, kwargs)

    agg_method = "cov" if demean else "mean"
    cov_: Callable = getattr(agg_buffer, agg_method)

    return annualisation * cov_()



"""Tradingo analytics functions."""

__all__ = [
    "cov",
    "expectation",
    "sharpe_ratio",
    "omega_ratio",
]

from .cov import cov
from .expectation import expectation
from .ratios import omega_ratio, sharpe_ratio



import os
import pathlib

import pytest

from tradingo import dag
from tradingo.settings import TradingoConfig


@pytest.mark.skip
def test_config():
    env = TradingoConfig.from_env(
        env={
            "TP_CONFIG_HOME": "/home/rory/dev/tradingo-plat/config/tradingo/",
            "TP_ARCTIC_URI": "lmdb:///home/rory/dev/tradingo-plat/data/prod/tradingo.db",
        }
    )
    env.to_env()

    out = dag.read_config_template(
        pathlib.Path("/home/rory/dev/tradingo-plat/config/tradingo/yfinance.yaml"),
        variables=os.environ,
    )

    dag.DAG.from_config(out)



import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal

from tradingo.analytics import cov

STDS = np.array([0.1, 0.2, 0.15])
CORRELATIONS = np.array(
    [
        [1.0, 0.3, 0.2],
        [0.3, 1.0, 0.4],
        [0.2, 0.4, 1.0],
    ]
)


def excel_columns(size):
    """generate excel-like columns

    :param size: number of columns

    :returns: an array of column names
    """
    result = []
    for i in range(size):
        col = ""
        x = i
        while True:
            x, r = divmod(x, 26)
            col = chr(65 + r) + col
            if x == 0:
                break
            x -= 1
        result.append(col)
    return result


def cov_from_corr(
    correlation_mtx: np.ndarray, std_array: np.ndarray, deannualisation: int = 260
) -> np.ndarray:
    """
    covariance matrix from a std vector and corr matrix

    :param correlation_mtx: correlation matrix
    :param std_array: standard deviation array
    :param deannualisation: de-annualise covariance (e.g. from annual to daily)
    """
    assert len(correlation_mtx.shape) == 2
    assert all(cs == std_array.size for cs in correlation_mtx.shape)
    assert (correlation_mtx == correlation_mtx.T).all()
    cov_matrix = np.outer(std_array, std_array) * correlation_mtx
    return cov_matrix / deannualisation


def correlated_returns(
    size: int,
    cov_mtx: np.ndarray,
    target_vol: float = 0.1,
    mean: float = 0,
    seed: int = 1234,
) -> pd.DataFrame:
    """
    generate plausible market returns

    :param size: sample size
    :param cov_mtx: covariance matrix, with appropriate annualization
    :param target_vol: target vol, with appropriate annualization
    :param mean: mean of the sample
    :param seed: random seed

    :returns: the correlated returns dataframe
    """

    vol_scale = (target_vol**2) / np.sqrt(cov_mtx.sum().sum())
    cov_mtx = vol_scale * cov_mtx
    mean = mean or [0 for _ in range(len(cov_mtx))]

    gen = np.random.RandomState(seed)
    smp = gen.multivariate_normal(mean, cov_mtx, size)

    index = pd.DatetimeIndex(pd.bdate_range(pd.Timestamp(2020, 1, 2), periods=size))
    columns = excel_columns(len(cov_mtx))

    return pd.DataFrame(smp, index=index, columns=columns)


def test_cov():
    columns = ["A", "B", "C"]
    cov_mtx = cov_from_corr(CORRELATIONS, STDS, deannualisation=260)
    df = correlated_returns(1000, cov_mtx, target_vol=0.1)

    actual_ewm = cov(df, how="ewm", halflife=100)
    expected_ewm = pd.DataFrame(
        [
            [0.00439475, 0.00281318, 0.00079727],
            [0.00281318, 0.01817437, 0.00548258],
            [0.00079727, 0.00548258, 0.0115796],
        ],
        index=pd.MultiIndex.from_product([[pd.Timestamp(2023, 11, 1)], columns]),
        columns=columns,
    )
    assert_frame_equal(actual_ewm.tail(3), expected_ewm)

    actual_rolling = cov(df, how="rolling", window=100)
    expected_rolling = pd.DataFrame(
        [
            [0.00421997, 0.00285151, 0.00144046],
            [0.00285151, 0.01600372, 0.00633751],
            [0.00144046, 0.00633751, 0.01300069],
        ],
        index=pd.MultiIndex.from_product([[pd.Timestamp(2023, 11, 1)], columns]),
        columns=columns,
    )
    assert_frame_equal(actual_rolling.tail(3), expected_rolling)

    actual_expanding = cov(df, how="expanding")
    expected_expanding = pd.DataFrame(
        [
            [0.00444099, 0.00259191, 0.00116196],
            [0.00259191, 0.01852227, 0.00611158],
            [0.00116196, 0.00611158, 0.01096141],
        ],
        index=pd.MultiIndex.from_product([[pd.Timestamp(2023, 11, 1)], columns]),
        columns=columns,
    )
    assert_frame_equal(actual_expanding.tail(3), expected_expanding)



"""Tradingo config."""

import json
import pathlib
from typing import Any

import jinja2
import yaml


class ConfigLoadError(Exception):
    """raised when loading config incorrectly."""


def read_config_template(filepath: pathlib.Path, variables) -> dict[str, Any]:
    """read a config template and populate it with values"""
    filepath = pathlib.Path(filepath)

    renderedtext = jinja2.Template(filepath.read_text(encoding="utf-8")).render(
        **variables
    )

    try:
        if filepath.suffix == ".json":
            return process_includes(json.loads(renderedtext), variables)

        if filepath.suffix == ".yaml":
            return process_includes(yaml.safe_load(renderedtext), variables)

    except jinja2.TemplateSyntaxError as ex:
        raise ConfigLoadError(f"Error reading config template: {filepath}") from ex

    raise ValueError(f"Unhandled file type: '{filepath.suffix}'")


def process_includes(config: dict[str, Any], variables) -> dict[str, Any]:
    """bake nested configs via 'include' pattern."""

    out = {}

    for key, value in config.items():
        if isinstance(value, dict) and "include" in value:
            value = process_includes(value, variables)

            protocol, path = value["include"].split("://")

            if protocol == "file":
                incvalue = read_config_template(
                    pathlib.Path(path),
                    variables={
                        **value.get("variables", {}),
                        **variables,
                    },
                )

            else:
                raise ValueError(
                    f"Unsupported protocol: '{protocol}' at '{key}' for '{path}'"
                )
            value = value.copy()
            value.update(incvalue)
            value.pop("include", None)
            value.pop("variables", None)

        elif isinstance(value, dict):
            value = process_includes(value, variables)

        out[key] = value

    return out



"""Expectation calculation module."""

from __future__ import annotations

from typing import Callable

import pandas as pd
from pandas.core.window.rolling import BaseWindow


def safe_dataframe_window(dataframe, how, kwargs) -> BaseWindow:
    """get a dataframe aggergation window safely"""
    try:
        aggregator: Callable = getattr(dataframe, how)
        agg_buffer: BaseWindow = aggregator(**kwargs)
        assert isinstance(agg_buffer, BaseWindow)
    except AttributeError as ex:
        raise NotImplementedError(f"how={how}") from ex
    except AssertionError as ex:
        raise ValueError(f"how={how} is not a dataframe BaseWindow aggregator") from ex
    return agg_buffer


def expectation(
    dataframe: pd.DataFrame,
    annualisation: int = 260,
    how: str = "exponential",
    **kwargs,
) -> pd.DataFrame:
    """
    Calculate the expectation of a given DataFrame.

    :param dataframe: DataFrame containing the data.
    :param annualisation: The annualisation factor, by default 260.
    :param how: The method of calculating expectation, either "exponential" or "rolling",
        by default "exponential".
    :param **kwargs: keyword arguments
        Additional arguments for the ewm (exponentially weighted mean) method.
        For rolling expectation, the window size must be specified in kwargs.
        For example, `window=10` for a 10-period rolling expectation.
        For exponential expectation, the halflife can be specified in kwargs.
        For example, `halflife=10` for a 10-period halflife.

    :returns: The expectation of the data.
    """

    agg_buffer = safe_dataframe_window(dataframe, how, kwargs)
    return annualisation * agg_buffer.mean()
